@@ -1,5 +1,3 @@
-from __future__ import division
-
 __copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"
 
 __license__ = """
@@ -24,30 +22,26 @@ THE SOFTWARE.
 
 import math
 import pymbolic
-from pymbolic.mapper.stringifier import (StringifyMapper, PREC_NONE,
+from pymbolic.mapper.stringifier import (
+        StringifyMapper, PREC_NONE,
         PREC_SUM, PREC_POWER)
 
 
-def _constant_mapper(c):
-    # work around numpy bug #1137 (locale-sensitive repr)
-    # http://projects.scipy.org/numpy/ticket/1137
-    try:
-        import numpy
-    except ImportError:
-        pass
-    else:
-        if isinstance(c, numpy.floating):
-            c = float(c)
-        elif isinstance(c, numpy.complexfloating):
-            c = complex(c)
-
-    return repr(c)
-
-
 class CompileMapper(StringifyMapper):
-    def __init__(self):
-        StringifyMapper.__init__(self,
-                constant_mapper=_constant_mapper)
+    def map_constant(self, expr, enclosing_prec):
+        # work around numpy bug #1137 (locale-sensitive repr)
+        # http://projects.scipy.org/numpy/ticket/1137
+        try:
+            import numpy
+        except ImportError:
+            pass
+        else:
+            if isinstance(expr, numpy.floating):
+                expr = float(expr)
+            elif isinstance(expr, numpy.complexfloating):
+                expr = complex(expr)
+
+        return repr(expr)
 
     def map_polynomial(self, expr, enclosing_prec):
         # Use Horner's scheme to evaluate the polynomial
@@ -58,9 +52,9 @@ class CompileMapper(StringifyMapper):
             if exp == 0:
                 return ""
             elif exp == 1:
-                return "*%s" % sbase
+                return f"*{sbase}"
             else:
-                return "*%s**%s" % (sbase, exp)
+                return f"*{sbase}**{exp}"
 
         result = ""
         rev_data = expr.data[::-1]
@@ -69,13 +63,11 @@ class CompileMapper(StringifyMapper):
                 next_exp = rev_data[i+1][0]
             else:
                 next_exp = 0
-            result = "(%s+%s)%s" % (result, self(coeff, PREC_SUM),
+            result = "({}+{}){}".format(result, self(coeff, PREC_SUM),
                     stringify_exp(exp-next_exp))
-        #print "A", result
-        #print "B", expr
 
         if enclosing_prec > PREC_SUM and len(expr.data) > 1:
-            return "(%s)" % result
+            return f"({result})"
         else:
             return result
 
@@ -87,9 +79,9 @@ class CompileMapper(StringifyMapper):
             else:
                 rec = stringify_leading_dimension
 
-            return "[%s]" % (", ".join(rec(x) for x in ary))
+            return "[{}]".format(", ".join(rec(x) for x in ary))
 
-        return "numpy.array(%s)" % stringify_leading_dimension(expr)
+        return "numpy.array({})".format(stringify_leading_dimension(expr))
 
     def map_foreign(self, expr, enclosing_prec):
         return StringifyMapper.map_foreign(self, expr, enclosing_prec)
@@ -107,15 +99,14 @@ class CompiledExpression:
         :arg variables: The first arguments (as strings or
             :class:`pymbolic.primitives.Variable` instances) to be used for the
             compiled function.  All variables used by the expression and not
-            present here are added in alphabetical order.
+            present here are added in lexicographic order.
         """
-        import pymbolic.primitives as primi
+        self._compile(expression, variables)
 
+    def _compile(self, expression, variables):
+        import pymbolic.primitives as primi
         self._Expression = expression
         self._Variables = [primi.make_variable(v) for v in variables]
-        self._compile()
-
-    def _compile(self):
         ctx = self.context().copy()
 
         try:
@@ -129,24 +120,24 @@ class CompiledExpression:
         used_variables = DependencyMapper(
                 composite_leaves=False)(self._Expression)
         used_variables -= set(self._Variables)
-        used_variables -= set(pymbolic.var(key) for key in ctx.keys())
+        used_variables -= {pymbolic.var(key) for key in list(ctx.keys())}
         used_variables = list(used_variables)
         used_variables.sort()
         all_variables = self._Variables + used_variables
 
         expr_s = CompileMapper()(self._Expression, PREC_NONE)
-        func_s = "lambda %s: %s" % (",".join(str(v) for v in all_variables),
+        func_s = "lambda {}: {}".format(",".join(str(v) for v in all_variables),
                 expr_s)
-        self.__call__ = eval(func_s, ctx)
-
-    def __getinitargs__(self):
-        return self._Expression, self._Variables
+        self._code = eval(func_s, ctx)
 
     def __getstate__(self):
-        return None
+        return self._Expression, self._Variables
 
     def __setstate__(self, state):
-        pass
+        self._compile(*state)
+
+    def __call__(self, *args):
+        return self._code(*args)
 
     def context(self):
         return {"math": math}
