@@ -1,5 +1,3 @@
-from __future__ import division, absolute_import
-
 __copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"
 
 __license__ = """
@@ -22,10 +20,9 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
+from sys import intern
 import pymbolic.traits as traits
 
-import six
-from six.moves import range, zip, intern
 
 __doc__ = """
 Expression base class
@@ -155,7 +152,7 @@ Helper functions
 Interaction with :mod:`numpy` arrays
 ------------------------------------
 
-:mod:`numpy.ndarray` instances are supported anywhere in an expression.
+:class:`numpy.ndarray` instances are supported anywhere in an expression.
 In particular, :mod:`numpy` object arrays are useful for capturing
 vectors and matrices of :mod:`pymbolic` objects.
 
@@ -174,7 +171,7 @@ def disable_subscript_by_getitem():
     pass
 
 
-class Expression(object):
+class Expression:
     """Superclass for parts of a mathematical expression. Overrides operators
     to implicitly construct :class:`Sum`, :class:`Product` and other expressions.
 
@@ -191,15 +188,17 @@ class Expression(object):
 
     .. method:: __getitem__
 
+    .. method:: __getinitargs__
+
     .. automethod:: make_stringifier
 
     .. automethod:: __eq__
+    .. automethod:: is_equal
     .. automethod:: __hash__
+    .. automethod:: get_hash
     .. automethod:: __str__
     .. automethod:: __repr__
 
-    .. automethod:: __repr__
-
     .. rubric:: Logical operator constructors
 
     .. automethod:: not_
@@ -377,7 +376,7 @@ class Expression(object):
 
     # {{{ bitwise operators
 
-    def __inv__(self):
+    def __invert__(self):
         return BitwiseNot(self)
 
     def __or__(self, other):
@@ -433,7 +432,7 @@ class Expression(object):
     def a(self):
         """Provide a spelling ``expr.a.name`` for encoding attribute lookup.
         """
-        class AttributeLookupCreator(object):
+        class AttributeLookupCreator:
             def __init__(self, aggregate):
                 self.aggregate = aggregate
 
@@ -455,29 +454,11 @@ class Expression(object):
             stringifier should carry forward attributes and settings of
             *originating_stringifier*.
         """
-        if originating_stringifier is None:
-            stringify_mapper_args = ()
-        else:
-            stringify_mapper_args = (originating_stringifier.constant_mapper,)
-
-        try:
-            stringifier_class_getter = self.stringifier
-        except AttributeError:
-            from pymbolic.mapper.stringifier import StringifyMapper
-            return StringifyMapper(*stringify_mapper_args)
-        else:
-            from warnings import warn
-            warn("%s overrides 'stringifier', which is deprecated. "
-                    "Override 'make_stringifier' instead. "
-                    "Backward compatibility will go away "
-                    "sometime after June 2020."
-                    % type(self).__name__,
-                    DeprecationWarning)
-
-            return stringifier_class_getter()(*stringify_mapper_args)
+        from pymbolic.mapper.stringifier import StringifyMapper
+        return StringifyMapper()
 
     def __str__(self):
-        """Use the :meth:`stringifier` to return a human-readable
+        """Use the :meth:`make_stringifier` to return a human-readable
         string representation of *self*.
         """
 
@@ -492,7 +473,7 @@ class Expression(object):
             if isinstance(child, tuple):
                 # Make sure limit propagates at least through tuples
 
-                return "(%s%s)" % (
+                return "({}{})".format(
                         ", ".join(strify_child(i, limit-1) for i in child),
                         "," if len(child) == 1 else "")
 
@@ -505,7 +486,7 @@ class Expression(object):
                 strify_child(i, limit-1)
                 for i in self.__getinitargs__())
 
-        return "%s(%s)" % (self.__class__.__name__, initargs_str)
+        return f"{self.__class__.__name__}({initargs_str})"
 
     def __repr__(self):
         """Provides a default :func:`repr` based on
@@ -660,6 +641,9 @@ class Expression(object):
 
     # }}}
 
+    def __abs__(self):
+        return Call(Variable("abs"), (self,))
+
     def __iter__(self):
         # prevent infinite loops (e.g. when inserting into numpy arrays)
         raise TypeError("expression types are not iterable")
@@ -698,7 +682,7 @@ class Variable(Leaf):
             return NotImplemented
 
     def __setstate__(self, val):
-        super(Variable, self).__setstate__(val)
+        super().__setstate__(val)
 
         self.name = intern(self.name)
 
@@ -753,9 +737,9 @@ class Call(AlgebraicLeaf):
             pass
         else:
             if len(self.parameters) != arg_count:
-                raise TypeError("%s called with wrong number of arguments "
-                        "(need %d, got %d)" % (
-                            self.function, arg_count, len(parameters)))
+                raise TypeError(
+                        f"{self.function} called with wrong number of arguments "
+                        f"(need {arg_count}, got {len(parameters)})")
 
     def __getinitargs__(self):
         return self.function, self.parameters
@@ -800,9 +784,9 @@ class CallWithKwargs(AlgebraicLeaf):
             pass
         else:
             if len(self.parameters) != arg_count:
-                raise TypeError("%s called with wrong number of arguments "
-                        "(need %d, got %d)" % (
-                            self.function, arg_count, len(parameters)))
+                raise TypeError(
+                        f"{self.function} called with wrong number of arguments "
+                        f"(need {arg_count}, got {len(parameters)})")
 
     def __getinitargs__(self):
         return (self.function,
@@ -1158,10 +1142,31 @@ class Comparison(Expression):
 
     init_arg_names = ("left", "operator", "right")
 
+    operator_to_name = {
+            "==": "eq",
+            "!=": "ne",
+            ">=": "ge",
+            ">": "gt",
+            "<=": "le",
+            "<": "lt",
+            }
+    name_to_operator = {name: op for op, name in operator_to_name.items()}
+
     def __init__(self, left, operator, right):
+        """
+        :arg operator: accepts the same values as :attr:`operator`, or the
+            standard Python comparison operator names
+
+        .. versionchanged:: 2020.2
+
+            Now also accepts Python operator names.
+        """
         self.left = left
         self.right = right
-        if operator not in [">", ">=", "==", "!=", "<", "<="]:
+
+        operator = self.name_to_operator.get(operator, operator)
+
+        if operator not in self.operator_to_name:
             raise RuntimeError("invalid operator")
         self.operator = operator
 
@@ -1467,7 +1472,7 @@ class Slice(Expression):
 
     @property
     def start(self):
-        if len(self.children) > 1:
+        if len(self.children) > 0:
             return self.children[0]
         else:
             return None
@@ -1590,14 +1595,11 @@ def quotient(numerator, denominator):
 global VALID_CONSTANT_CLASSES
 global VALID_OPERANDS
 VALID_CONSTANT_CLASSES = (int, float, complex)
-if six.PY2:
-    VALID_CONSTANT_CLASSES += (long,)  # noqa pylint:disable=undefined-variable
-
 VALID_OPERANDS = (Expression,)
 
 try:
     import numpy
-    VALID_CONSTANT_CLASSES += (numpy.number,)
+    VALID_CONSTANT_CLASSES += (numpy.number, numpy.bool_)
 except ImportError:
     pass
 
@@ -1670,19 +1672,22 @@ def make_common_subexpression(field, prefix=None, scope=None):
         return field
 
     try:
-        from pytools.obj_array import log_shape
+        import numpy
+        have_obj_array = (
+            isinstance(field, numpy.ndarray)
+            and field.dtype.char == "O")
+        logical_shape = (
+            field.shape
+            if isinstance(field, numpy.ndarray)
+            else ())
     except ImportError:
         have_obj_array = False
-    else:
-        have_obj_array = True
-
-    if have_obj_array:
-        ls = log_shape(field)
+        logical_shape = ()
 
     from pymbolic.geometric_algebra import MultiVector
     if isinstance(field, MultiVector):
         new_data = {}
-        for bits, coeff in six.iteritems(field.data):
+        for bits, coeff in field.data.items():
             if prefix is not None:
                 blade_str = field.space.blade_bits_to_str(bits, "")
                 component_prefix = prefix+"_"+blade_str
@@ -1694,11 +1699,9 @@ def make_common_subexpression(field, prefix=None, scope=None):
 
         return MultiVector(new_data, field.space)
 
-    elif have_obj_array and ls != ():
-        from pytools import indices_in_shape
-        result = numpy.zeros(ls, dtype=object)
-
-        for i in indices_in_shape(ls):
+    elif have_obj_array and logical_shape != ():
+        result = numpy.zeros(logical_shape, dtype=object)
+        for i in numpy.ndindex(logical_shape):
             if prefix is not None:
                 component_prefix = prefix+"_".join(str(i_i) for i_i in i)
             else:
@@ -1711,6 +1714,7 @@ def make_common_subexpression(field, prefix=None, scope=None):
                         field[i], component_prefix, scope)
 
         return result
+
     else:
         if is_constant(field):
             return field
@@ -1738,9 +1742,9 @@ def make_sym_vector(name, components, var_factory=Variable):
     if isinstance(components, Integral):
         components = list(range(components))
 
-    from pytools.obj_array import join_fields
+    from pytools.obj_array import flat_obj_array
     vfld = var_factory(name)
-    return join_fields(*[vfld.index(i) for i in components])
+    return flat_obj_array(*[vfld.index(i) for i in components])
 
 
 def make_sym_array(name, shape, var_factory=Variable):
@@ -1750,8 +1754,7 @@ def make_sym_array(name, shape, var_factory=Variable):
 
     import numpy as np
     result = np.zeros(shape, dtype=object)
-    from pytools import indices_in_shape
-    for i in indices_in_shape(shape):
+    for i in np.ndindex(shape):
         result[i] = vfld.index(i)
 
     return result
